/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file SubAssembler.h */

#ifndef AMDIS_SUB_ASSEMBLER_H
#define AMDIS_SUB_ASSEMBLER_H

#include <map>
#include <vector>
#include "AMDiS_fwd.h"
#include "FixVec.h"
#include "Flag.h"
#include <boost/any.hpp>

namespace AMDiS {

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Base class for SecondOrderAssembler, FirstOrderAssembler, 
   * ZeroOrderAssembler. The task of a SubAssembler is to assemble a list of 
   * terms of a special order and add their contributions to a DOFMatrix or a 
   * DOFVector. An Assembler can consist of up to four SubAssemblers: one 
   * SecondOrderAssembler for second order terms, one ZeroOrderAssembler for 
   * terms of order zero, and two FirstOrderAssemblers. One for terms with
   * derivatives of the basis functions connected to to row DOFs and one for 
   * those connected to the column DOFs.
   */
  class SubAssembler
  {
  public:
    /// Creates a SubAssembler belonging to assembler for the terms of given 
    /// order of Operator op. If order is equal to one, type spezifies what kind 
    /// of FirstOrderType are to assemble. During construction of a SubAssembler
    /// the needs and properties of the terms are considered.
    SubAssembler(Operator *op,
		 Assembler *assembler,
		 Quadrature *quadrat,
		 int order, 
		 bool optimized,
		 FirstOrderType type = GRD_PHI);

    /// Destructor
    virtual ~SubAssembler() {}

    /// Calculates the element matrix for elInfo and adds it to mat. Memory for
    /// mat must be provided by the caller.
    virtual void calculateElementMatrix(const ElInfo *elInfo,
					ElementMatrix& mat) = 0;

    /// Calculates the element vector for elInfo and adds it to vec. Memory for
    /// vec must be provided by the caller.
    virtual void calculateElementVector(const ElInfo *elInfo,
					ElementVector& vec) = 0;

    /// Returns \ref terms
    inline std::vector<OperatorTerm*> *getTerms() 
    { 
      return &terms; 
    }

    /// Returns \ref quadrature.
    inline Quadrature *getQuadrature() 
    {
      return quadrature;
    }

    /// Sets \ref quadrature to q.
    inline void setQuadrature(Quadrature* q) 
    {
      quadrature = q;
    }
  
    /// Creates a vector with the world coordinates of the quadrature points
    /// of \ref quadrature on the element of elInfo. 
    /// Used by \ref OperatorTerm::initElement().
    void getCoordsAtQPs(const ElInfo* elInfo, 
			Quadrature *quad,
			mtl::dense_vector<WorldVector<double> >& coordsAtQPs);

    /// DOFVector dv evaluated at quadrature points.
    /// Used by \ref OperatorTerm::initElement().
    template<typename T>
    void getVectorAtQPs(DOFVectorBase<T>* dv, 
			const ElInfo* elInfo,
			Quadrature *quad,
			mtl::dense_vector<T>& vecAtQPs);

    ///
    template<typename T>
    void getVectorAtQPs(DOFVectorBase<T>* dv, 
			const ElInfo* smallElInfo,
			const ElInfo* largeElInfo,
			Quadrature *quad,
			mtl::dense_vector<T>& vecAtQPs);
    
    /// Gradients of DOFVector dv evaluated at quadrature points.
    /// Used by \ref OperatorTerm::initElement().
    template<typename T>
    void getGradientsAtQPs(DOFVectorBase<T>* dv,
			   const ElInfo* elInfo,
			   Quadrature *quad,
			   mtl::dense_vector<typename GradientType<T>::type>& grdAtQPs);

    template<typename T>
    void getGradientsAtQPs(DOFVectorBase<T>* dv,
			   const ElInfo* smallElInfo,
			   const ElInfo* largeElInfo,
			   Quadrature *quad,
			   mtl::dense_vector<typename GradientType<T>::type>& grdAtQPs);

    /// The comp'th component of the derivative of DOFVector dv evaluated at
    /// quadrature points. Used by \ref OperatorTerm::initElement().
    /// Attention: not caching at the moment! Using cache if gradients for read 
    /// but not for write.
    template<typename T>
    void getDerivativeAtQPs(DOFVectorBase<T>* dv,
			    const ElInfo* elInfo,
			    Quadrature *quad,
			    int comp,
			    mtl::dense_vector<T>& grdAtQPs);

    template<typename T>
    void getDerivativeAtQPs(DOFVectorBase<T>* dv,
			    const ElInfo* smallElInfo,
			    const ElInfo* largeElInfo,
			    Quadrature *quad,
			    int comp,
			    mtl::dense_vector<T>& grdAtQPs);

    /// Called once for each ElInfo when \ref calculateElementMatrix() or
    /// \ref calculateElementVector() is called for the first time for this
    /// Element.
    virtual void initElement(const ElInfo *smallElInfo,
			     const ElInfo *largeElInfo = NULL,
			     Quadrature *quad = NULL);

    /// Returns \ref psiFast.
    const FastQuadrature *getPsiFast() const 
    { 
      return psiFast; 
    }

    // Returns \ref phiFast.
    const FastQuadrature *getPhiFast() const 
    { 
      return phiFast; 
    }

    /// Returns \ref name.
    std::string getName() const
    {
      return name;
    }

  protected:
    /// Updates \ref psiFast and \ref phiFast.
    FastQuadrature *updateFastQuadrature(FastQuadrature *quadFast,
					 const BasisFunction *psi,
					 Flag updateFlag);
  
  protected:
    /// Problem dimension
    int dim;

    /// Row FiniteElemSpace.
    const FiniteElemSpace *rowFeSpace;

    /// Column FiniteElemSpace.
    const FiniteElemSpace *colFeSpace;

    /// Number of rows of the element matrix and length of the element
    /// vector. Is equal to the number of row basis functions
    int nRow;

    /// Number of columns of the element matrix. Is equal to the number
    /// of column basis functions
    int nCol;

    /// Used for \ref getVectorAtQPs() and \ref getGradientsAtQPs().
    struct ValuesAtQPs {
      ValuesAtQPs()
	: valid(false), quad(NULL)
      {}
      ValuesAtQPs(boost::any values_, bool valid_, Quadrature* quad_=NULL)
	: values(values_), valid(valid_), quad(quad_)
      {}
      
      boost::any values; // used for mtl::dense_vector<T>
      bool valid;
      Quadrature* quad;
    };

    std::map<const void*, ValuesAtQPs* > cachedValuesAtQPs;
    std::map<const void*, ValuesAtQPs* > cachedGradientsAtQPs;
    
    /// Set and updated by \ref initElement() for each ElInfo. 
    /// coordsAtQPs[i] points to the coordinates of the i-th quadrature point.
    mtl::dense_vector<WorldVector<double> > cacheCoordsAtQPs;

    /// Used for \ref getCoordsAtQPs().
    bool coordsValid;

    /// Used for \ref getCoordsAtQP(). Stores the number of allocated 
    /// WorldVectors.
    int coordsNumAllocated;

    /// Quadrature object to be used for assembling.
    Quadrature *quadrature;

    /// FastQuadrature for row basis functions
    FastQuadrature *psiFast;

    /// FastQuadrature for column basis functions
    FastQuadrature *phiFast;

    /// Flag that specifies whether the element matrix is symmetric.
    bool symmetric;

    /// List of all terms with a contribution to this SubAssembler
    std::vector<OperatorTerm*> terms;

    ///
    bool opt;

    ///
    bool firstCall;

    /// Name of the assembler. Is used to print information about 
    /// used assembler.
    std::string name;

    friend class Assembler;
  };

}

#include "SubAssembler.hh"

#endif // AMDSIS_SUB_ASSEMBLER_H
